package serial
// Serial io as raw as possible. 8 bits, no parity 1 stop bit, channel to provide DTR control
//
// comName := "com3" 
// baudRate := 57600
// debug := false
//
// chanFromSerial, chanToSerial, charDTRSerial, chanQuitSerial, err := SerialChannels( comName, baud, debug)
//
// chanFromSerial = byte data channel from serial
// chanToSerial - byte data channel to serial 
// chanDTRSerial - bool channel to control DTR - logical 1 means DTR active - rs232 level - +3-+15 volts 
// chanQuitSerial - send a bit on this channel to terminate
//
// based on http://code.google.com/p/goserial/
//
// Works with go version 1, tested on Windows 7(64 bit) and xp (32bit)
import (
	"errors"
	"fmt"
	"os"
	"sync"
	"syscall"
	"time"
	"unsafe"
)

type serialPort struct { //**********PBP //**********PBP
	fp  *os.File //**********PBP
	fd syscall.Handle //**********PBP
	rl sync.Mutex //**********PBP
	wl sync.Mutex //**********PBP
	ro *syscall.Overlapped //**********PBP
	wo *syscall.Overlapped //**********PBP
} //**********PBP

type structDCB struct {
	DCBlength, BaudRate uint32
	flags [4]byte
	wReserved, XonLim, XoffLim uint16
	ByteSize, Parity, StopBits byte
	XonChar, XoffChar, ErrorChar, EofChar, EvtChar byte
	wReserved1 uint16
}

type structTimeouts struct {
	ReadIntervalTimeout uint32
	ReadTotalTimeoutMultiplier uint32
	ReadTotalTimeoutConstant uint32
	WriteTotalTimeoutMultiplier uint32
	WriteTotalTimeoutConstant uint32
}

func SerialChannels( name string, baud int, conLog chan string, debug bool) (chan byte, chan byte, chan bool, chan bool, chan bool, error) { //**********PBP
	p := new(serialPort) //**********PBP
	var err error //**********PBP

	chFromSerial := make(chan byte,8192)
	chToSerial := make(chan byte,8192)
	chQuitSerial := make(chan bool)
	chDTRSerial := make(chan bool)
	chQuitSerial1 := make(chan bool) //**********PBP
	chQuitSerial2 := make(chan bool) //**********PBP
	
	if len(name) >0 && name[0] != '\\' {
		name = "\\\\.\\" + name
	}
	p.fd, err = syscall.CreateFile( syscall.StringToUTF16Ptr(name), //**********PBP
		syscall.GENERIC_READ|syscall.GENERIC_WRITE, 0, nil, syscall.OPEN_EXISTING,
		syscall.FILE_ATTRIBUTE_NORMAL|syscall.FILE_FLAG_OVERLAPPED, 0)
	if err == nil {
		p.fp = os.NewFile( uintptr(p.fd), name) //**********PBP
	}
	if err == nil {
		err = setCommState(p.fd, baud) //**********PBP
	}
	if err == nil {
		err = setupComm(p.fd, 64, 64) //**********PBP
	}
	if err == nil {
		err = setCommTimeouts(p.fd) //**********PBP
	}
	if err == nil {
		err = setCommMask(p.fd) //**********PBP
	}
	if err == nil {
		p.ro, err = newOverlapped() //**********PBP
	}
	if err == nil {
		p.wo, err = newOverlapped() //**********PBP
	}
	if  err != nil {
		err = errors.New( "OpenPort ERROR: " + err.Error())
	} else {

		chFromSerial = make( chan byte, 8192)
		chQuitFrom := make( chan bool)
		go func() {
			var count int
			var done uint32
			then := time.Now()
			datain := make( []byte, 8192)
			for q := false; q == false; {
				p.rl.Lock() //**********PBP
				readerr := resetEvent(p.ro.HEvent) //**********PBP
				if readerr == nil {
					readerr = syscall.ReadFile(p.fd , datain, &done, p.ro) //**********PBP
					if readerr == syscall.ERROR_IO_PENDING {
						count, readerr = getOverlappedResult(p.fd, p.ro) //**********PBP
					} else {
						count = int(done)
					}
				}
				p.rl.Unlock() //**********PBP
				
				din := datain[:count]

				if debug && len(din) > 0 {
					now := time.Now()
					ds1 := fmt.Sprintf("(%v(%d)", now.Sub(then), len(din))
					then = now
					for _, c := range( din) {
						if c < byte(0x20) || c > byte(0x7E) {
							ds1 += fmt.Sprintf("{%02X}",c)
						} else {
							ds1 += fmt.Sprintf("%c", c)
						}
					}
					conLog <- ds1+")\n"
				}
				for _, c := range( din) {
					chFromSerial <- c
				}
				select {
  				case q = <-chQuitFrom:
				default:
					time.Sleep(10) //**********PBP
				}
			}
			chQuitSerial1 <- true //**********PBP
		}()

		chToSerial = make( chan byte, 8192)
		chQuitTo := make( chan bool)
		const fBuffer int = 8192 //**********PBP
		go func() {
			dataout := make( []byte, fBuffer) //**********PBP
			var count, wcount int
			var readflag bool
			then := time.Now()
			for q := false; q == false; {
				select {
				case q = <- chQuitTo:
				case dataout[ 0] = <- chToSerial:
					for readflag, wcount = true, 1; readflag && wcount < fBuffer; wcount++ { //**********PBP
						select {
						case dataout[wcount] = <- chToSerial:
						default:
							readflag = false
							wcount--
						}
					}

					dout := dataout[:wcount]
					
					if debug {
						now := time.Now()
						ds1 := fmt.Sprintf("\t[%v[%d]", now.Sub(then), len(dout))
						then = now
						for _, c := range( dout) {
							if c < byte(0x20) || c > byte(0x7E) {
								ds1 += fmt.Sprintf("{%02X}",c)
							} else {
								ds1 += fmt.Sprintf("%c", c)
							}
						}
						conLog <- ds1+"]\n"
					}
					p.wl.Lock() //**********PBP
					writeerr := resetEvent(p.wo.HEvent) //**********PBP
					if  writeerr == nil {
						var n uint32 = uint32( wcount)
						writeerr = syscall.WriteFile( p.fd, dout, &n, p.wo) //**********PBP
						count = int( n)
						if writeerr == syscall.ERROR_IO_PENDING {
							for i := 0; i < 10; i++ {
								count, writeerr = getOverlappedResult(p.fd, p.wo) //**********PBP
								if count != wcount {
									conLog <- fmt.Sprintf( "Serial port write erro3 %d %s %v %v %v\n", i,  writeerr, writeerr, count, wcount)
									time.Sleep( 1e9) //**********PBP
								} else {
									break
								}
							}
						}
					}
					p.wl.Unlock() //**********PBP
					if count != wcount {
						conLog <- fmt.Sprintf( "Serial port write error1 %s %v %v %v\n", writeerr, writeerr, count, wcount)
					}
				}
			}
			chQuitSerial1 <- true //**********PBP
		}()

		chQuitDTR := make( chan bool)
		go func() {
			for q := false; q == false; {
				select {
				case q = <- chQuitDTR:
				case dtr := <-chDTRSerial:
					if dtr {
						escapeCommFunction( p.fd, 5)
					} else {
						escapeCommFunction( p.fd, 6)
					}
				}
			}
			chQuitSerial1 <- true //**********PBP
		}()
		
		go func() {
			<- chQuitSerial
			p.fp.Close() //**********PBP
			chQuitFrom <- true
			<- chQuitSerial1 //**********PBP
			chQuitTo <- true
			<- chQuitSerial1 //**********PBP
			chQuitDTR <- true
			<- chQuitSerial1 //**********PBP
			chQuitSerial2 <- true //**********PBP
		}()
 	}
	return chFromSerial, chToSerial, chDTRSerial, chQuitSerial, chQuitSerial2, err //**********PBP
} // end SerialChannels

var (
	nSetCommState,
	nEscapeCommFunction,
	nSetCommTimeouts,
	nSetCommMask,
	nSetupComm,
	nGetOverlappedResult,
	nCreateEvent,
	nResetEvent uintptr
)

func init() {
	handle, err := syscall.LoadLibrary("kernel32.dll")
	if err == nil {
		nSetCommState = getProcAddr(handle, "SetCommState")
		nEscapeCommFunction = getProcAddr(handle, "EscapeCommFunction")
		nSetCommTimeouts = getProcAddr(handle, "SetCommTimeouts")
		nSetCommMask = getProcAddr(handle, "SetCommMask")
		nSetupComm = getProcAddr(handle, "SetupComm")
		nGetOverlappedResult = getProcAddr(handle, "GetOverlappedResult")
		nCreateEvent = getProcAddr(handle, "CreateEventW")
		nResetEvent = getProcAddr(handle, "ResetEvent")
		syscall.FreeLibrary(handle)
	} else {
		panic("LoadLibrary " + err.Error())
	}
}

func getProcAddr(handle syscall.Handle, name string) uintptr {
	addr, err := syscall.GetProcAddress(handle, name)
	if err != nil {
		panic(name + " " + err.Error())
	}
	return addr
}

func escapeCommFunction(handle syscall.Handle, fparam int) error {
	r, _, err := syscall.Syscall(nEscapeCommFunction, 2, uintptr(handle), uintptr(fparam), 0)
	if r == 0 {
		return err
	}
	return nil
}

func setCommState(handle syscall.Handle, baud int) error {
	var params structDCB
	params.DCBlength = uint32(unsafe.Sizeof(params))
	params.flags[0] = 0x01 // fBinary
	params.BaudRate = uint32(baud)
	params.ByteSize = 8
	r, _, err := syscall.Syscall(nSetCommState, 2, uintptr(handle), uintptr(unsafe.Pointer(&params)), 0)
	if r == 0 {
		return err
	}
	return nil
}
//
// timeout set to be as short as possible
//
func setCommTimeouts(handle syscall.Handle) error {
	var timeouts structTimeouts
	const MAXDWORD = 1<<32 - 1
	timeouts.ReadIntervalTimeout = MAXDWORD
	timeouts.ReadTotalTimeoutMultiplier = 0
	timeouts.ReadTotalTimeoutConstant =  0
	timeouts.WriteTotalTimeoutMultiplier = MAXDWORD //1**********PBP
	timeouts.WriteTotalTimeoutConstant = 0
	r, _, err := syscall.Syscall(nSetCommTimeouts, 2, uintptr(handle), uintptr(unsafe.Pointer(&timeouts)), 0)
	if r == 0 {
		return err
	}
	return nil
}

func setupComm(handle syscall.Handle, in, out int) error {
	r, _, err := syscall.Syscall(nSetupComm, 3, uintptr(handle), uintptr(in), uintptr(out))
	if r == 0 {
		return err
	}
	return nil
}

func setCommMask(handle syscall.Handle) error {
	const EV_RXCHAR = 0x0001
	r, _, err := syscall.Syscall(nSetCommMask, 2, uintptr(handle), EV_RXCHAR, 0)
	if r == 0 {
		return err
	}
	return nil
}

func resetEvent(handle syscall.Handle) error {
	r, _, err := syscall.Syscall(nResetEvent, 1, uintptr(handle), 0, 0)
	if r == 0 {
		return err
	}
	return nil
}

func newOverlapped() (*syscall.Overlapped, error) {
	var overlapped syscall.Overlapped
	r, _, err := syscall.Syscall6(nCreateEvent, 4, 0, 1, 0, 0, 0, 0)
	if r == 0 {
		return nil, err
	}
	overlapped.HEvent = syscall.Handle(r)
	return &overlapped, nil
}

func getOverlappedResult(handle syscall.Handle, overlapped *syscall.Overlapped) (int, error) {
	var n int
	r, _, err := syscall.Syscall6(nGetOverlappedResult, 4,
	uintptr(handle),
	uintptr(unsafe.Pointer(overlapped)),
	uintptr(unsafe.Pointer(&n)), 1, 0, 0)
	if r == 0 {
		return n, err
	}
	return n, nil
}
